import os
import numpy as np
import torch
import shutil
import torchvision.transforms as transforms
import torchvision.datasets as dset
from torch.autograd import Variable
import wandb


class AvgrageMeter(object):

  def __init__(self):
    self.reset()

  def reset(self):
    self.avg = 0
    self.sum = 0
    self.cnt = 0

  def update(self, val, n=1):
    self.sum += val * n
    self.cnt += n
    self.avg = self.sum / self.cnt


@torch.no_grad()
def accuracy(output, target, topk=(1,)):
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(k=maxk, dim=1, largest=True, sorted=True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(dim=0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))

    return res


class Cutout(object):
    def __init__(self, length):
        self.length = length

    def __call__(self, img):
        h, w = img.size(1), img.size(2)
        mask = np.ones((h, w), np.float32)
        y = np.random.randint(h)
        x = np.random.randint(w)

        y1 = np.clip(y - self.length // 2, 0, h)
        y2 = np.clip(y + self.length // 2, 0, h)
        x1 = np.clip(x - self.length // 2, 0, w)
        x2 = np.clip(x + self.length // 2, 0, w)

        mask[y1: y2, x1: x2] = 0.
        mask = torch.from_numpy(mask)
        mask = mask.expand_as(img)
        img *= mask
        return img


def drop_path(x, drop_prob):
    if drop_prob > 0.:
        keep_prob = 1. - drop_prob
        mask = Variable(torch.cuda.FloatTensor(
            x.size(0), 1, 1, 1).bernoulli_(keep_prob))
        x.div_(keep_prob)
        x.mul_(mask)
    return x
    

def _data_transforms_cifar10(args):
  CIFAR_MEAN = [0.49139968, 0.48215827, 0.44653124]
  CIFAR_STD = [0.24703233, 0.24348505, 0.26158768]

  train_transform = transforms.Compose([
    transforms.RandomCrop(32, padding=4),
    transforms.RandomHorizontalFlip(),
    transforms.ToTensor(),
    transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
  ])
  if args.cutout:
    train_transform.transforms.append(Cutout(args.cutout_length))

  valid_transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
    ])
  return train_transform, valid_transform


def _data_transforms_cifar10_eval(args):
  train_transform = transforms.Compose([
    transforms.RandomCrop(32, padding=4),
    transforms.RandomHorizontalFlip(),
    transforms.ToTensor(),
  ])
  if args.cutout:
    train_transform.transforms.append(Cutout(args.cutout_length))

  valid_transform = transforms.Compose([
    transforms.ToTensor()
    ])
  return train_transform, valid_transform


def _data_transforms_cifar100(args):
  CIFAR_MEAN = [0.5071, 0.4867, 0.4408]
  CIFAR_STD = [0.2675, 0.2565, 0.2761]

  train_transform = transforms.Compose([
    transforms.RandomCrop(32, padding=4),
    transforms.RandomHorizontalFlip(),
    transforms.ToTensor(),
    transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
  ])
  if args.cutout:
    train_transform.transforms.append(Cutout(args.cutout_length))

  valid_transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
    ])
  return train_transform, valid_transform


def _data_imagenet(args):
  traindir = os.path.join(args.data, 'train')
  #validdir = os.path.join(args.data, 'val')
  normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
  train_data = dset.ImageFolder(
    traindir,
    transforms.Compose([
      transforms.RandomResizedCrop(224),
      transforms.RandomHorizontalFlip(),
      transforms.ToTensor(),
      normalize,
    ]))
  return train_data


def count_parameters_in_MB(model):
  return np.sum(np.prod(v.size()) for name, v in model.named_parameters() if "auxiliary" not in name)/1e6


def save_checkpoint(state, is_best, save, epoch):
  filename = os.path.join(save, 'checkpoint-epoch{}.pth.tar'.format(epoch))
  
  if is_best:
    best_filename = os.path.join(save, 'model_best.pth.tar')
    torch.save(state, best_filename)
  else:
    torch.save(state, filename)


def save(model, model_path):
  torch.save(model.state_dict(), model_path)


def load(model, model_path):
  model.load_state_dict(torch.load(model_path))


def drop_path(x, drop_prob):
  if drop_prob > 0.:
    keep_prob = 1.-drop_prob
    mask = Variable(torch.cuda.FloatTensor(x.size(0), 1, 1, 1).bernoulli_(keep_prob))
    x.div_(keep_prob)
    x.mul_(mask)
  return x


def create_exp_dir(path, scripts_to_save=None):
  if not os.path.exists(path):
    os.mkdir(path)
  print('Experiment dir : {}'.format(path))

  if scripts_to_save is not None:
    os.mkdir(os.path.join(path, 'scripts'))
    for script in scripts_to_save:
      dst_file = os.path.join(path, 'scripts', os.path.basename(script))
      shutil.copyfile(script, dst_file)


class Logger:
    def __init__(
        self,
        exp_name,
        log_dir=None,
        exp_suffix="",
        write_textfile=True,
        use_wandb=False,
        wandb_project_name=None,
        entity='sh'
    ):

        self.log_dir = log_dir
        self.write_textfile = write_textfile
        self.use_wandb = use_wandb

        self.logs_for_save_clean = {}
        self.logs_clean = {}
        
        self.logs_for_save_adv = {}
        self.logs_adv = {}

        if self.write_textfile:
            self.f_clean = open(os.path.join(log_dir, 'logs_clean.txt'), 'w')
            self.f_adv = open(os.path.join(log_dir, 'logs_adv.txt'), 'w')

        if self.use_wandb:
            exp_suffix = "_".join(exp_suffix.split("/")[:-1])
            self.run = wandb.init(
                config=wandb.config,
                entity=entity,
                project=wandb_project_name, 
                name=exp_name + "_" + exp_suffix, 
                group=exp_name.split('net-')[0],
                reinit=True)
            
            
    def update_config(self, v, is_args=False, is_clean=False):
        if is_args:
          if is_clean:
            self.logs_for_save_clean.update({'args': v})
          else:
            self.logs_for_save_adv.update({'args': v})
        else:
          if is_clean:
            self.logs_for_save_clean.update(v)
          else:
            self.logs_for_save_adv.update(v)
        if self.use_wandb:
            wandb.config.update(v, allow_val_change=True)


    def write_log_nohead(self, element, step, is_clean=False):
        log_str = f"{step} | "
        log_dict = {}
        if is_clean: 
          logs_for_save = self.logs_for_save_clean
          f = self.f_clean
        else:
          logs_for_save = self.logs_for_save_adv
          f = self.f_adv
          
        for key, val in element.items():
            if not key in logs_for_save:
                logs_for_save[key] =  []
            logs_for_save[key].append(val)
            log_str += f'{key} {val} | '
            log_dict[f'{key}'] = val
        
        if self.write_textfile:
            f.write(log_str+'\n')
            f.flush()

        if self.use_wandb:
            wandb.log(log_dict, step=step)
      
            
    def write_log(self, element, step, img_dict=None, tbl_dict=None):
        log_str = f"{step} | "
        log_dict = {}
        
        if is_clean: 
          logs_for_save = self.logs_for_save_clean
          logs = self.logs_clean
          f = self.f_clean
        else:
          logs_for_save = self.logs_for_save_adv
          logs = self.logs_adv
          f = self.f_adv
          
        for head, keys  in element.items():
            for k in keys:
                v = logs[k].avg
                if not k in logs_for_save:
                    logs_for_save[k] = []
                logs_for_save[k].append(v)
                log_str += f'{k} {v}| '
                log_dict[f'{head}/{k}'] = v

        if self.write_textfile:
            f.write(log_str+'\n')
            f.flush()
        
        if img_dict is not None:
            log_dict.update(img_dict)
        
        if tbl_dict is not None:
            log_dict.update(tbl_dict)
            
        if self.use_wandb:
            wandb.log(log_dict, step=step)


    def save_log(self, name=None):
        torch.save(self.logs_for_save_clean, os.path.join(self.log_dir, 'logs_clean.pt'))
        torch.save(self.logs_for_save_adv, os.path.join(self.log_dir, 'logs_adv.pt'))
    

    def update(self, key, v, n=1, is_clean=False):
        logs = self.logs_clean if is_clean else self.logs_adv
        if not key in logs:
            logs[key] = AverageMeter()
        logs[key].update(v, n)
    

    def reset(self, keys=None, except_keys=[], is_clean=False):
        logs = self.logs_clean if is_clean else self.logs_adv
        if keys is not None:
            if isinstance(keys, list):
                for key in keys:
                    logs[key] =  AverageMeter()
            else:
                logs[keys] = AverageMeter()
        else:
            for key in logs.keys():
                if not key in except_keys:
                    logs[key] = AverageMeter()


    def avg(self, keys=None, except_keys=[], is_clean=False):
        logs = self.logs_clean if is_clean else self.logs_adv
        if keys is not None:
            if isinstance(keys, list):
                return {key: logs[key].avg for key in keys if key in logs.keys()}
            else:
                return logs[keys].avg
        else:
            avg_dict = {}
            for key in logs.keys():
                if not key in except_keys:
                    avg_dict[key] =  logs[key].avg
            return avg_dict 


